#!/usr/bin/python3

# hello there

# this is a program designed for the transfer of files over
# the gopher protocol. it uses cgi scripts and server queries
# to accomplish this

# to install this program, modify the variable dbfile to the
# place you would like to store your database, and then use
# the unix touch command to create a file in its place.
# without the touch command the server will fail with error
# code 2.

# after that, you must symlink the main file to the files
# 'getid', 'post', 'finish', and 'get', and chmod +x the main
# code.

# finally, run `chown nobody` on your dbfile if you're running
# gophernicus or the like

# should you encounter an error, here's a catalogue of server-
# side errors:

# - error code 1: unknown invocation (you symlinked to the
# wrong file)
# - error code 2: cannot open database (try `touch`ing the
# dbfile
# - error code 3: cannot write database (make sure the euid
# has proper permissions to the database)

# this program is licenced under the GNU AGPL 3

import base64
import os
import sys

dbfile = "/var/gopher/cgi-bin/dig/digdb"

class GopherDigError(BaseException): pass
class FinishedError(GopherDigError): pass
class IdNotFoundError(GopherDigError): pass

class GFile:
    def __init__(self, line, id=None, finished=None, content=None):
        if not line == None:
            try:
                splitted = line.split(' ')
                if splitted[0][0] == "*": self.finished = True
                else: self.finished = False
                self.id = int(splitted[0].split('*')[-1])
                if len(splitted) == 2:
                    self.content = splitted[1]
                else:
                    self.content = ""
            except:
                raise GopherDigError('error while parsing line')
        else:
            self.id = id
            self.finished = finished
            self.content = content

    def write(self, fd):
        try:
            if self.finished: fd.write("*")
            fd.write(str(self.id) + ' ' + self.content + '\n')
        except BaseException as e:
            printer(e)
    def append(self, data):
        self.content += data
    def setfinished(self, data=None):
        if data == None: return self.finished
        else: self.finished = data
    def returndecoded(self):
        return base64.b64decode(self.content)

class DB:
    def __init__(self, file):
        self.children = []
        self.max = 0
        with open(file, "r") as fd:
            for i in fd.readlines():
                i = i.rstrip()
                self.children.append(GFile(i))
                if self.children[-1].id > self.max:
                    self.max = self.children[-1].id
    def write(self, file):
        with open(file, "w") as fd:
            for i in self.children:
                i.write(fd)
    def newid(self):
        self.max += 1
        self.children.append(GFile(None, self.max, False, ""))
        return self.max
    def _findid(self, id):
        for i in self.children:
            if i.id == id: return i
        raise IdNotFoundError('id not found')
    def _append(self, id, data):
        self._findid(id).append(data)
    def append(self, line):
        splitted = line.split(' ')
        id = int(splitted[0])
        if not self._findid(id).finished:
            self._append(id, splitted[1])
        else:
            raise FinishedError("item is finished")
    def setfinished(self, id):
        self._findid(id).setfinished(True)
    def returndecoded(self, id):
        return self._findid(id).returndecoded()

def getid(d, i, j):
    id = d.newid()
    printer(str(id))

def post(d, envs, i):
    try: d.append(envs.get("QUERY_STRING"))
    except FinishedError:
        printer("ERR: FINISHED")
        return
    except IdNotFoundError:
        printer("ERR: UNK ID")
        return
    except:
        printer("ERR: UNK ERR")
        return
    printer("APPENDED")

def finish(d, envs, i):
    try: d.setfinished(int(envs.get("QUERY_STRING")))
    except:
        printer("ERR: UNK ERR")
        return
    printer("FINISHED")

def get(d, envs, i):
    try: decoded = d.returndecoded(int(envs.get("QUERY_STRING")))
    except:
        printer("ERR: UNK ERR")
        return
    sys.stdout.buffer.write(decoded)

def active(d, envs, i):
    printer("ACTIVE")

def printer(data):
    sys.stdout.write(data)

def main():
    running = sys.argv[0]
    if not len(running) > 2:
        running = os.environ.get("REQUEST").split('/')[-1]#.rstrip()
    try: d = DB(dbfile)
    except: printer("e2"); sys.exit(2)
    envs = os.environ
    if "getid" in running: getid(d, envs, sys.argv[1:])
    elif "post" in running: post(d, envs, sys.argv[1:])
    elif "finish" in running: finish(d, envs, sys.argv[1:])
    elif "get" in running: get(d, envs, sys.argv[1:])
    elif "active" in running: active(d, envs, sys.argv[1:])
    else: printer("e1"); sys.exit(1)
    try: d.write(dbfile)
    except: printer("e3"); sys.exit(3)

if __name__ == "__main__": main()
